from typing import Literal, TypeAlias, Sequence, Mapping
from pydantic import BaseModel

Result = Literal["1-0", "draw", "0-1", "+-", "-+"]

class Paired(BaseModel):
    white: str
    black: str
    white_no: int | None = None
    white_elo: int | None = None
    black_no: int | None = None
    black_elo: int | None = None
    result: Result | None = None
    paired: bool = True
    
class Unpaired(BaseModel):
    player: str
    reason: str
    paired: bool = False
    
Pairing: TypeAlias = Paired | Unpaired
RoundPairings: TypeAlias = Sequence[Pairing]
GroupPairings: TypeAlias = Sequence[RoundPairings]
TournamentPairings: TypeAlias = Mapping[str, GroupPairings]
"""Group -> Round -> Board -> Pairing"""


from .errors import DownloadError, ParsingError, ScrapingError
from ._donwload import download, download_main, download_pairings, download_round_pairings, download_schedule
from .parsing import parse_result, parse_columns, parse_row, parse_rounds
from .main import scrape_pairings

__all__ = [
  'DownloadError', 'ParsingError', 'ScrapingError',
  'download', 'download_main', 'download_pairings', 'download_round_pairings', 'download_schedule',
  'parse_result', 'parse_columns', 'parse_row', 'parse_rounds',
  'scrape_pairings'
]


from typing import Literal, TypeAlias
from dataclasses import dataclass

@dataclass
class DownloadError:
  detail: str | None = None
  reason: Literal['download-error'] = 'download-error'

@dataclass
class ParsingError:
  detail: str | None = None
  reason: Literal['parsing-error'] = 'parsing-error'

ScrapingError: TypeAlias = DownloadError | ParsingError


from typing import Iterable
from dataclasses import dataclass
import re
from bs4 import BeautifulSoup, Tag
from haskellian import Either, Left, Right, IsLeft, iter as I, Iter, either as E
from chess_pairings import Result, Paired, Unpaired, Pairing
from .errors import ParsingError

def parse_result(result: str) -> Result | None:
  """Parse result string from chessresults page
  - `'1 - 0' -->` white victory
  - `'0 - 1' -->` black victory
  - `'"½ - ½"' -->` draw
  - `'+ - -' -->` white victory by forfeit
  - `'- - +' -->` black victory by forfeit
  - any other --> None
  """
  match result:
    case "1 - 0":
      return "1-0"
    case "0 - 1":
      return "0-1"
    case "½ - ½":
      return "draw"
    case "+ - -":
      return "+-"
    case "- - +":
      return "-+"

@dataclass
class Columns:
  white: int
  black: int
  result: int
  white_no: int | None = None
  white_elo: int | None = None
  black_no: int | None = None
  black_elo: int | None = None

@dataclass
class Row:
  white: str
  black: str
  result: str
  white_no: str | None = None
  white_elo: str | None = None
  black_no: str | None = None
  black_elo: str | None = None

def parse_columns(soup: BeautifulSoup) -> Either[ParsingError, Columns]:
  for row in soup.find_all('tr'):
    try:
      headers = [th.get_text(strip=True) for th in row.find_all("th", recursive=False)]
      if headers != []:
        white = I.find_idx(lambda x: x == "Name", headers)
        black = I.find_last_idx(lambda x: x == "Name", headers)
        result = I.find_idx(lambda x: x == "Result", headers)
        white_no = I.find_idx(lambda x: x == "No.", headers)
        white_elo = I.find_idx(lambda x: x == "Rtg", headers)
        black_no = I.find_last_idx(lambda x: x == "No.", headers)
        black_elo = I.find_last_idx(lambda x: x == "Rtg", headers)
        if white is not None and black is not None and result is not None:
          return Right(Columns(white, black, result, white_no, white_elo, black_no, black_elo))
    except:
      ...
  return Left(ParsingError('Unable to find headers'))

@I.lift
def extract_round(table: Tag, columns: Columns) -> Iterable[Row]:
  rows = table.find_all('tr', class_=["CRng1", "CRng2", "CRg1", "CRg2"])
  for row in rows:
    try:
      cols = list(row.find_all("td"))
      col_text = lambda i: cols[i].get_text(strip=True)
      yield Row(
        white=col_text(columns.white), black=col_text(columns.black), result=col_text(columns.result),
        white_no=col_text(columns.white_no) if columns.white_no is not None else None,
        white_elo=col_text(columns.white_elo) if columns.white_elo is not None else None,
        black_no=col_text(columns.black_no) if columns.black_no is not None else None,
        black_elo=col_text(columns.black_elo) if columns.black_elo is not None else None,
      )
    except:
      ...

def safe_int(s) -> int | None:
  return E.safe(lambda: int(s)).get_or(None)

def parse_row(row: Row) -> Pairing:
  if row.black_no == '':
    return Unpaired(player=row.white, reason=row.black)
  else:
    return Paired(
      result=parse_result(row.result), white=row.white, black=row.black,
      white_no=safe_int(row.white_no), white_elo=safe_int(row.white_elo),
      black_no=safe_int(row.black_no), black_elo=safe_int(row.black_elo),
    )
  
def parse_rounds(soup: BeautifulSoup) -> Either[ParsingError, list[list[Pairing]]]:
  try:
    rounds: dict[int, list[Pairing]] = {}
    columns = parse_columns(soup).unsafe()
    headings = soup.find_all(string=re.compile("^Round ."))

    for h in headings:
      rnd = int(h.get_text(strip=True).split(" ")[1])
      table = h.find_next("table")
      pairs = extract_round(table, columns).map(parse_row).sync()
      if any(isinstance(p, Paired) for p in pairs):
        rounds[rnd] = pairs

    sorted_rounds = Iter(rounds.items()).sort(lambda kv: kv[0]).map(lambda kv: kv[1]).sync()
    return Right(sorted_rounds)

  except IsLeft as e:
    return Left(e.value)


from .types import Result, Paired, Unpaired, Pairing, RoundPairings, GroupPairings, TournamentPairings
from .chess_results import scrape_pairings
from . import chess_results

__all__ = [
  'Result', 'Paired', 'Unpaired', 'Pairing', 'RoundPairings', 'GroupPairings', 'TournamentPairings',
  'chess_results', 'scrape_pairings',
]


from haskellian import Either
from chess_pairings import GroupPairings
from . import download_pairings, parse_rounds, ScrapingError

async def scrape_pairings(db_key: int) -> Either[ScrapingError, GroupPairings]:
  return (await download_pairings(db_key)).bind(parse_rounds)


from aiohttp import ClientSession
from bs4 import BeautifulSoup
from haskellian import Either, Left, Right, either as E
from .errors import DownloadError

DB_KEY = 741703
def slug(db_key: int) -> str:
    return f"/tnr{db_key}.aspx"

CHESS_RESULTS = "https://chess-results.com/"
BLOCKED_MESSAGE = "Note: To reduce the server load by daily scanning of all links (daily 100.000 sites and more) by search engines like Google, Yahoo and Co, all links for tournaments older than 5 days (end-date) are shown after clicking the following button:"

async def download(
  slug: str, params: dict = {}, *,
  base = CHESS_RESULTS,
  blocked_message = BLOCKED_MESSAGE
) -> Either[DownloadError, BeautifulSoup]:
  """Download any chess-results page bypassing the anti-indexing block"""
  async with ClientSession(base) as ses:
    r = await ses.get(slug, params=params)
    page = await r.content.read()
    soup = BeautifulSoup(page, 'html.parser')
    if soup.find(string=blocked_message) is None:
      return Right(soup)
    else:
      input_tag = soup.find(id="__VIEWSTATE")
      if input_tag is None:
        return Left(DownloadError('VIEWSTATE not found'))
      viewstate = E.safe(lambda: input_tag['value']).get_or(None) # type: ignore
      if viewstate is None:
        return Left(DownloadError('VIEWSTATE found but does not have "value" attribute'))
      r2 = await ses.post(slug, params=params, data={
        "__VIEWSTATE": viewstate,
        "cb_alleDetails": "Show+tournament+details"
      })
      page2 = await r2.content.read()
      return Right(BeautifulSoup(page2, 'html.parser'))
    
async def download_main(db_key: int):
    """Main page of a tournament. Bypasses anti-indexing block"""
    url = slug(db_key)
    params = dict(lan=1)
    return await download(url, params)

async def download_round_pairings(db_key: int, round: int):
    """Pairings page. Bypasses anti-indexing block"""
    params = dict(lan=1, rd=round, art=2, turdet="YES")
    url = slug(db_key)
    return await download(url, params)

async def download_pairings(db_key: int):
    params = dict(lan=1, rd=0, art=2, turdet="YES", zeilen=99999)
    url = slug(db_key)
    return await download(url, params)

async def download_schedule(db_key: int):
    params = dict(lan=1, art=14, turdet="YES")
    url = slug(db_key)
    return await download(url, params)